#
#   Copyright (c) 2014-2015 eGauge Systems LLC
# 	4730 Walnut St, Suite 110
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
default_app_config = 'epic.apps.EPIC_App_Config'



#!/usr/bin/env python
#
#   Copyright (c) 2014-2018 eGauge Systems LLC
#	4730 Walnut St, Suite 110
#	Boulder, CO 80301
#	voice: 720-545-9767
#	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
'''This is a KiCad Eeschema plugin that converts a KiCad XML to:

	1) a bill-of-materials (BOM), output as CSV text file
	   (comma-separated values)
	2) an EPIC part (assembly) that consists of the BOM components

The EPIC assembly created by this tool will have its manufacturer set
to django.conf.EPIC_MANUFACTURER and the manufacturer's part-number
will be set to 'bom:' with the name of the schematic and flavor-name
(if any) appended.  If an assembly with that part-number already
exists, it will be updated unless the BOM was edited manually inside
EPIC.  In the latter case, you can force overwriting the manually
edited BOM pass passing this script the --force-overwrite command-line
option.

This tool also updates the Eeschema schematic files with the footprint
info in the EPIC database.  The original schematic files are backed up
in files with a tilde (~) appended to the original filename.  If the
footprint in a schematic file changes, the tool will warn you about
that and remind you to re-generate the net list file and potentially
update the PCB.

This tool extracts part number an installation information from the
XML file using the following component field names:

	Field Name:	Purpose:
	------------	---------------------------------------------
	EP		Specifies the EPIC part number to use for the
			component.

	Installation	If the value of this field is "DNP" (Do Not Place),
			then the component is omitted from the BOM.
			In the CSV output file, do-not-place components
			are listed separately at the end of the file.

A single Eeschema file may define multiple flavors (versions) of a
BOM.  This is done by appending the flavor name to the above field
names, separated by a dash (-).  For example, to define a flavor
called `lite', you'd use field names `EP-lite' and/or
`Installation-lite' when a particular component needs to have a
different EPIC part number or a different installation method for that
flavor.  Use command-line option --flavor to specify which flavor to
generate the a BOM for.

'''

from __future__ import print_function

from argparse import RawDescriptionHelpFormatter as RawDescHelpFormatter
from types import SimpleNamespace

import argparse
import io
import os
import re
import sys
import tempfile
import xml.etree.ElementTree as ET

# Please replace 'epic-sample' with the name of your Django project:
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'epic-sample.epic-sample.settings')

import django
django.setup()

from django.conf import settings as cfg
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from epic.models import Part, format_part_number, Assembly_Item, strchoice, \
    Vendor

from os import path

MY_TOOL_NAME = 'kicad-to-epic-bom'

manufacturer = cfg.EPIC_MANUFACTURER

class Component:
    def __init__(self, part_id=None, refdes=None, value=None, footprint=None,
                 mfg=None, mfg_pn=None, part=None):
        self.part_id = part_id
        self.part = part
        self.refdes = refdes
        self.value = value
        self.footprint = footprint
        self.mfg = mfg
        self.mfg_pn = mfg_pn

    def __str__(self):
        res = 'Component('
        sep = ''
        for f in dir(self):
            if f[0] == '_':
                continue
            val = self.__dict__[f]
            if val != None:
                res += '%s%s=%s' % (sep, f, val)
                sep = ','
        res += ')'
        return res

def quote(str):
    str = str.replace('"', '\"')
    # Replace comma by colon.  DigiKey doesn't
    # handle commas correctly even if they're inside double quotes...
    str = str.replace(',', ':')
    return '"' + str + '"'

@transaction.atomic
def save_assembly(assembly_part, bom):
    created = assembly_part.id is None

    assembly_part.last_bom_mod_type = Part.LAST_MOD_TYPE_TOOL
    assembly_part.last_bom_mod_name = MY_TOOL_NAME
    assembly_part.save()

    Assembly_Item.objects.filter(assy_id=assembly_part.id).delete()

    for components in bom:
        comp = components[0]
        refdes = ','.join(c.refdes for c in components)
        item = Assembly_Item(assy_id=assembly_part.id,
                             comp_id=comp.part_id,
                             qty=len(components),
                             refdes=refdes)
        item.save()
    print('%s: info: %s part %s (%s %s) with %d components' \
          % (MY_TOOL_NAME, 'created' if created else 'updated',
             format_part_number(assembly_part.id),
             assembly_part.mfg, assembly_part.mfg_pn, len(bom)),
          file=sys.stderr)

def update_comp(comp):
    footprint_field_index = part_number = None
    has_changed = False
    refdes = ''
    for index, line in enumerate(comp):
        m = re.match(r'F\s+(\d+)\s+(\S+)\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+'
                     '\s+\S+(\s+(\S+))?$', line)
        if m:
            if m.group(1) == '0':
                if m.lastindex >= 2:
                    refdes = m.group(2)[1:-1] + ' '
            elif m.group(1) == '2':
                footprint_field_index = index
            elif m.lastindex >= 3 and m.group(4) == '"EP"':
                try:
                    part_str = m.group(2)[1:-1]
                    part_number = int(part_str)
                except ValueError:
                    print('%s: warn: invalid EPIC part number "%s"' %
                          (MY_TOOL_NAME, part_str), file=sys.stderr)
    if footprint_field_index is not None and part_number is not None:
        try:
            part = Part.objects.get(pk=part_number)
        except ObjectDoesNotExist:
            print('%s: error: Part %s does not exist' %
                  (MY_TOOL_NAME, format_part_number(part_number)),
                  file=sys.stderr)
            return False
        if part.footprint is None or part.footprint == '':
            if part.mounting != Part.MOUNTING_CHASSIS \
               and part.mounting != Part.MOUNTING_FREE:
                print('%s: warn: EPIC part %s has no footprint defined' %
                      (MY_TOOL_NAME, format_part_number(part.id)),
                      file=sys.stderr)
        else:
            old = comp[footprint_field_index]
            m = re.match(r'F\s+2\s+"([^"]*)"+\s+(.*)', old)
            if m:
                old_footprint = m.group(1)
                rest = m.group(2)
                new_footprint = part.footprint
                new = 'F 2 \"%s\" %s\n' % (new_footprint, rest)
                comp[footprint_field_index] = new
                if old_footprint != new_footprint:
                    print('%s: warn: %sfootprint changed from %s to %s.'
                          % (MY_TOOL_NAME, refdes, old_footprint,
                             new_footprint))
                    has_changed = True
                else:
                    has_changed = False
    return has_changed

def update_schematic(dirname, filename):
    '''Update schematic file FILENAME residing in directory DIRNAME with
    footprints from the EPIC database.

    '''
    schem_path = os.path.join(dirname, filename)
    has_changes = False
    with tempfile.NamedTemporaryFile(dir=dirname) as out:
        with io.open(schem_path, 'r', encoding='utf-8') as schem_file:
            inside_comp = False
            for line in schem_file:
                if line.rstrip() == '$Comp':
                    inside_comp = True
                    curr_comp = [line]
                elif line.rstrip() == '$EndComp':
                    inside_comp = False
                    curr_comp.append(line)
                    has_changes |= update_comp(curr_comp)
                    for comp_line in curr_comp:
                        out.write(comp_line.encode('utf-8'))
                elif inside_comp:
                    curr_comp.append(line)
                else:
                    out.write(line.encode('utf-8'))
        if has_changes:
            os.rename(schem_path, schem_path + '~')
            os.link(os.path.join(dirname, out.name), schem_path)
            print('%s: info: %s updated.  Use kicad to update net list and PCB'
                  % (MY_TOOL_NAME, schem_path))

def bom_append(bom, component):
    if component.part_id not in bom:
        bom[component.part_id] = []
    bom[component.part_id].append(component)

def bom_sort(bom):
    bom_list = []
    for part_id in bom:
        bom_list.append(bom[part_id])
    bom_list.sort(key=lambda x: x[0].part_id)
    return bom_list

def output_csv(outfile, bom_list, with_vendor_pn=False, preferred_vendors=None):
    for components in bom_list:
        qty = len(components)
        refdes = ','.join(c.refdes for c in components)
        c = components[0]

        substitutes = ''
        if c.part:
            sep = ''
            for sub in c.part.equivalent_parts():
                if sub == c.part:
                    continue
                substitutes += sep
                substitutes += '%s (%s %s)' % (format_part_number(sub.id),
                                               sub.mfg, sub.mfg_pn)
                sep = ', '


        row = '%s,%s,%s,%s,%s,%s,%s,%s' % \
              (qty, c.value, c.footprint, format_part_number(c.part_id),
               c.mfg, quote(c.mfg_pn), quote(substitutes), quote(refdes))

        if with_vendor_pn:
            vp = c.part.best_vendor_part(preferred_vendors)
            if vp is None:
                vendor = 'n/a'
                vendor_pn = 'n/a'
            else:
                vendor = vp.vendor
                vendor_pn = vp.vendor_pn
            row += ',%s,%s' % (vendor, vendor_pn)
        print(row, file=outfile)

def update(res, field, desired_flavor, detected_flavors):
    '''Process RES by matching FIELD.attrib['name'] against RES.PATTERN.
    If there is a match M, M.group(2) must evaluate to the name of the
    flavor of the field.  If this flavor matches the specified flavor,
    we have an exact match and RES.value is set to field.text, RES.name
    is set to field.attrib['name'].
    If the flavor of the field is empty and RES.value is None we have a
    default match and RES is updated like for an exact match.
    DETECTED_FLAVOR is a set of flavors found.
    '''
    m = res.pattern.match(field.attrib['name'])
    if m is None:
        return

    this_flavor = None
    if m.lastindex is not None:
        this_flavor = m.group(2)
        detected_flavors |= {this_flavor} # update set of all detected flavors

    if this_flavor == desired_flavor \
       or (this_flavor is None and res.value is None):
        res.name = field.attrib['name']
        res.value = field.text

def main():
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=RawDescHelpFormatter)
    parser.add_argument('input-file', nargs=1, help='Name of XML file '
                        'generated by KiCad\'s \"Generate BOM\" command.')
    parser.add_argument('output-file', nargs=1, help='Name of output file '
                        'that will contain the bill-of-materials (BOM) in '
                        'a comma-separated-values (CSV) file.')
    parser.add_argument('--flavor', nargs=1, default=None,
                        help='Select PCB flavor for which to generate BOM '
                        'for.')
    parser.add_argument('-f', '--force-overwrite', action='store_true',
                        help='Force overwriting of existing EPIC BOM '
                        '(assembly) part even if the part appears to '
                        'have been modified manually.')
    parser.add_argument('-V', '--with-vendor-pn', action='store_true',
                        help='Include Vendor and Vendor part-number in the '
                        'CSV output.')
    parser.add_argument ('-p', '--prefer-vendor', nargs='+', default=[],
                         help='Prefer the specified vendor(s), even if those '
                         'do not have the lowest cost for a given part. '
                         'Non-preferred vendors are used only for parts that '
                         'could be ordered otherwise.')
    parser.add_argument('-n', '--no-update-footprints', action='store_true',
                        help='By default, this program will update the '
                        'schematic files with the footprint information '
                        'found in the parts database.  This option can be '
                        'specified to skip this step.')
    parser.add_argument('--no-db-save', action='store_true',
                        help='Skip creating (or updating) the EPIC database '
                        'with the new BOM. '
                        'This is intended mainly for testing.')
    args = parser.parse_args()

    flavor = None
    if args.flavor is not None:
        flavor = args.flavor[0]

    preferred_vendors = None
    for vendor_name in args.prefer_vendor:
        vendors = Vendor.objects.filter(name__istartswith=vendor_name)
        if len(vendors) < 1:
            print('%s: error: vendor `%s\' cannot be found' %
                  (MY_TOOL_NAME, vendor_name), file=sys.stderr)
            sys.exit(1)
        elif len(vendors) > 1:
            print('%s: error: vendor name `%s\' not unique, matches: %s' %
                  (MY_TOOL_NAME, vendor_name,
                   ', '.join(['`%s\'' % v.name for v in vendors])),
                  file=sys.stderr)
            sys.exit(1)
        if preferred_vendors is None:
            preferred_vendors = []
        preferred_vendors.append(vendors[0])

    output_filename = args.__dict__['output-file'][0]
    input_filename = args.__dict__['input-file'][0]

    xml = ET.parse(input_filename)

    schematic_name = 'unknown'
    revision = ''
    design = xml.find('design')
    sources = []
    source_path = path.dirname(input_filename)
    if design is not None:
        source = design.find('source')
        if source is not None:
            schematic_name = path.splitext(path.basename(source.text))[0]
        sheet = design.find('sheet')
        for sheet in design.iter('sheet'):
            title_block = sheet.find('title_block')
            if title_block is not None:
                if revision == '':
                    rev = title_block.find('rev')
                    if rev is not None:
                        revision = '-rev' + rev.text.lstrip().rstrip()
                sources.append(title_block.find('source').text)

    if not args.no_update_footprints:
        for schematic in sources:
            update_schematic(source_path, schematic)

    assembly_name = 'bom:' + schematic_name + revision
    if flavor is not None:
        assembly_name += '-' + flavor
    qs = Part.objects.filter(mfg=manufacturer).filter(mfg_pn=assembly_name)
    if qs.exists():
        assembly_part = qs[0]
        if assembly_part.last_bom_mod_type != Part.LAST_MOD_TYPE_TOOL \
           or assembly_part.last_bom_mod_name != MY_TOOL_NAME:
            print('%s: info: part %s %s was last modified by %s %s' %
                  (MY_TOOL_NAME, manufacturer, assembly_name,
                   strchoice(Part.LAST_MOD_CHOICES,
                             assembly_part.last_bom_mod_type),
                   assembly_part.last_bom_mod_name), file=sys.stderr)
            if args.force_overwrite:
                print('%s: info: overwriting existing part due to '
                      '--force-overwrite'
                      % MY_TOOL_NAME, file=sys.stderr)
            else:
                print('%s: error: refusing to overwrite part %s; '
                      'use --force-overwrite if desired' \
                      % (MY_TOOL_NAME, format_part_number(assembly_part.id)),
                      file=sys.stderr)
                sys.exit(1)
    else:
        desc = 'BOM for %s' % schematic_name
        if flavor is not None:
            desc += (flavor + ' flavor')
        assembly_part = Part(descr=desc,
                             mfg=manufacturer, mfg_pn=assembly_name,
                             mounting=Part.MOUNTING_CHASSIS,
                             target_price=1000,
                             overage=1, spq=1, lead_time=4,
                             status=Part.STATUS_PREVIEW)

    ep_pat = re.compile(r'EP(-(.*))?')
    inst_pat = re.compile(r'Installation(-(.*))?')

    bom = {}    # regular bom
    dnp = {}    # do-not-place bom
    # part number & refdes used to refer to a particular part:
    src_part_dict = {}
    detected_flavors = set()
    for comp in xml.find('components'):
        refdes = comp.attrib.get('ref')

        part = SimpleNamespace(name=None, value=None, pattern=ep_pat)
        inst = SimpleNamespace(name=None, value=None, pattern=inst_pat)
        fields = comp.find('fields')
        if fields is not None:
            for field in fields:
                update(part, field, flavor, detected_flavors)
                update(inst, field, flavor, detected_flavors)

        do_not_place = (inst.value == 'DNP')

        if part.value is None:
            if not do_not_place:
                print('%s: warn: %s skipped due to missing '
                      'EPIC Part Number ("EP" field)'
                      % (MY_TOOL_NAME, refdes), file=sys.stderr)
            continue

        if do_not_place:
            print('%s: info: %s marked as do-not-place ("%s=DNP")'
                  % (MY_TOOL_NAME, refdes, inst.name), file=sys.stderr)

        try:
            part_id = int(part.value)
        except ValueError:
            print('%s: warn: %s has invalid EPIC part number "%s"' %
                  (MY_TOOL_NAME, refdes, part.value), file=sys.stderr)
            continue

        try:
            src_part = Part.objects.get(pk=part_id)
        except ObjectDoesNotExist:
            print('%s: warn: part %s not found in database' %
                  (MY_TOOL_NAME, format_part_number(part_id)), file=sys.stderr)
            continue

        c = Component(part_id, part=src_part, refdes=refdes,
                      value=comp.findtext('value', default='n/a'))
        best_part = src_part.best_part()

        # keep a list of part names used to refer to best part:
        if best_part.id not in src_part_dict:
            src_part_dict[best_part.id] = [src_part.id, c.refdes]
        elif src_part.id != src_part_dict[best_part.id][0]:
            print('%s: warn: %s uses %s instead '
                  'of equivalent %s used by %s' % \
                  (MY_TOOL_NAME, c.refdes,
                   format_part_number(src_part.id),
                   format_part_number(src_part_dict[best_part.id][0]),
                   src_part_dict[best_part.id][1]))

        if c.value != c.part.val:
            print('%s: warn: %s has value %s but part %s has value %s'
                  % (MY_TOOL_NAME, c.refdes, c.value,
                     format_part_number(part_id), c.part.val), file=sys.stderr)
        c.footprint = c.part.footprint
        c.mfg = c.part.mfg
        c.mfg_pn = c.part.mfg_pn

        if do_not_place:
            bom_append(dnp, c)
        else:
            bom_append(bom, c)

    if len(detected_flavors) == 0:
        detected_flavors_str = 'no flavors'
    else:
        if len(detected_flavors) > 1:
            detected_flavors_str = 'flavors: '
        else:
            detected_flavors_str = 'flavor: '
        detected_flavors_str += ', '.join(sorted(detected_flavors))

    if flavor is not None and flavor not in detected_flavors:
        print('%s: error: flavor `%s\' not defined by this schematic; '
              'detected %s'
              % (MY_TOOL_NAME, flavor, detected_flavors_str),
              file=sys.stderr)
        sys.exit(1)
    elif len(detected_flavors) > 0:
        print('%s: info: this schematic defines %s'
              % (MY_TOOL_NAME, detected_flavors_str), file=sys.stderr)

    bom_list = bom_sort(bom)
    dnp_list = bom_sort(dnp)

    out = io.open(output_filename, 'w', encoding='utf-8')
    hdr = (u'Qty,Value,Footprint,%s PN,Mfg,Mfg PN,Approved Substitutes,Refdes' %
           manufacturer)
    if args.with_vendor_pn:
        hdr += u',Vendor,Vendor PN'
    print(hdr, file=out)

    output_csv(out, bom_list, args.with_vendor_pn, preferred_vendors)

    if len(dnp_list) > 0:
        print('\nDO NOT PLACE parts:', file=out)
        output_csv(out, dnp_list, args.with_vendor_pn, preferred_vendors)

    if not args.no_db_save:
        save_assembly(assembly_part, bom_list)

if __name__ == '__main__':
    main()



#
#   Copyright (c) 2014-2015, 2017-2018 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
'''Read footprint info from KiCad footprints directory.'''
from datetime import datetime, timedelta

import os
import re

from django.conf import settings as cfg

class Footprints:
    last_update = datetime.now()
    list = None

    @classmethod
    def get(cls):
        if Footprints.list is None \
           or (datetime.now() - Footprints.last_update > timedelta(seconds=60)):
            Footprints.list = []
            pattern = re.compile(r'([.]pretty(/|$))')
            for root, _, files in os.walk(cfg.EPIC_KICAD_FOOTPRINTS_DIR):
                libname = pattern.sub(':',
                                      root[len(cfg.EPIC_KICAD_FOOTPRINTS_DIR):])
                for f in files:
                    m = re.match(r'(.*)[.]kicad_mod$', f)
                    if m:
                        Footprints.list.append(libname + m.group(1))
            Footprints.last_update = datetime.now()
        return Footprints.list



#
#   Copyright (c) 2018 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
import re

import tablib

from django.core.exceptions import ValidationError

from epic.models import Part, format_part_number

LINE_ITEM_COLUMNS = ['qty', 'part', 'amount']
DELTA_COLUMNS = ['qty', 'part']

ALTERNATE_NAMES = {
    'qty': ['quantity', 'count'],
    'part': ['pn', 'ep'],
    'amount': ['cost']
}

def clean_part(part):
    '''Verify that "part" is a valid (existing) part number.  May be
    specified as an number or as a string starting with the letters
    "EP".  Case is ignored.  For example: "5", "ep5", "EP0005" would
    all refer to part number 5.  Returns the part specified by
    "pn_str" or raises ValidationError in case of error.

    '''
    if isinstance(part, str):
        m = re.match(r'ep(\d+)$', part, re.IGNORECASE)
        if m:
            part = int(m.group(1))
        else:
            try:
                part = int(part)
            except ValueError:
                raise ValidationError('"%s" is not a valid part number.' %
                                      part)
    qs = Part.objects.filter(id=part)
    if not qs.exists():
        raise ValidationError('Part %s does not exist.' %
                              format_part_number(part))
    return qs[0]

def clean_qty(qty):
    '''Verify that "qty" is a valid quantity, which must be a non-negative
    number or a decimal integer string.  Returns the number or raises
    ValidationError in case of error.

    '''
    if isinstance(qty, str):
        try:
            qty = int(qty)
        except ValueError:
            raise ValidationError('"%s" is not a valid quantity.' % qty)
    if qty < 0:
        raise ValidationError('Quantity %d is invalid (must be non-negative).'
                              % qty)
    return qty

def clean_amount(amount):
    '''Verify that "amount" is a valid currency amount, which must be a
    non-negative number or decimal number string.  Returns the number
    or raises ValidationError in case of error.

    '''
    if isinstance(amount, str):
        try:
            amount = float(amount)
        except ValueError:
            raise ValidationError('"%s" is not a valid amount.' % amount)
    if amount < 0.0:
        raise ValidationError('Amount %f is invalid (must be non-negative).'
                              % amount)
    return amount

CLEANER = {
    'part': clean_part,
    'qty': clean_qty,
    'amount': clean_amount
}

def clean_spreadsheet(spreadsheet, required_columns, add_error):
    '''Given uploaded file "spreadsheet", verify that it has the columns
    specified by "required_columns" and return a list containing one
    entry for each data row in the spreadsheet.  Each list entry
    consists of dictionary with a name/value pair for each of the
    required columns.  Extraneous columns are ignored.  Errors that
    abort any further parsing of the spreadsheet are reported by
    raising ValidationError.  Other errors are reported by calls to
    "add_error", which must be a function expecting a ValidationError
    argument.

    '''
    contents = spreadsheet.read()
    try:
        # try loading utf-8 decoded contents for CSV files:
        txt = contents.decode('utf-8')
        ds = tablib.Dataset().load(txt)
    except Exception:
        try:
            # try loading binary contents for Excel files:
            ds = tablib.Dataset().load(contents)
        except Exception:
            raise ValidationError('failed to load spreadsheet; '
                                  'neither an CSV nor Excel file?')

    have_errors = False

    col_idx = {}	# maps required column to the index in the spreadsheet
    col_name = {}	# maps required column to the name in the spreadsheet
    for idx, name in enumerate(ds.headers):
        col = name.lower()
        for required in required_columns:
            if col == required or col in ALTERNATE_NAMES[required]:
                if required in col_idx:
                    add_error('Multiple columns for %s found (%s and %s)' %
                              (required, col_name[required], col))
                    have_errors = True
                col_idx[required] = idx
                col_name[required] = col
    for required in required_columns:
        if required not in col_idx:
            add_error(ValidationError(
                'Column %s missing.  It may also be called one of: %s.' %
                (required, ', '.join(ALTERNATE_NAMES[required]))))
            have_errors = True

    if have_errors:
        return []
    if not ds:
        raise ValidationError('At least one row of data required.')

    rows = []
    for row in ds:
        clean = {}
        for required in required_columns:
            try:
                val = CLEANER[required](row[col_idx[required]])
            except ValidationError as e:
                add_error(e)
            else:
                clean[required] = val
        rows.append(clean)
    return rows



#
#   Copyright (c) 2014-2015, 2018 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
'''Unit tests.  Run, e.g., from epic-sample with:

	python3 manage.py test epic.tests
'''
import io
import os
import re

from datetime import timedelta

from django.test import TestCase
from django.utils import timezone
from django.contrib.auth import get_user_model
from django.contrib.auth import models as auth_models

import epic
from epic import perms
from epic.compat import reverse
from epic.models import Assembly_Item, Delta, Inventory, Line_Item, Order, \
    Part, Shipment, Vendor, Vendor_Part, Warehouse
from epic import urls

class URL_Iterator:
    def __init__(self, urlpatterns):
        self.urlpattern_iterator = urlpatterns.__iter__()

    def __iter__(self):
        return self

    def __next__(self):
        # pylint: disable=protected-access
        try:
            pat = self.urlpattern_iterator.__next__() # Python 3
        except AttributeError:
            pat = self.urlpattern_iterator.next() # Python 2
        regex = re.compile(pat.pattern._regex)
        args = {}
        for key in regex.groupindex.keys():
            args[key] = 1
        url = reverse('epic:%s' % pat.name, kwargs=args)
        return url, pat.name

    def next(self):
        """For Python 2 compatibility."""
        return self.__next__()

def create_part(pk):
    return Part(pk=pk, mfg='mfg', mfg_pn='mfg_pn%d' % pk,
                mounting=Part.MOUNTING_SMD, target_price=1, overage=0,
                spq=1, status=Part.STATUS_ACTIVE, lead_time=9)

def create_warehouse(pk, name):
    return Warehouse(pk=pk, name=name)

def create_vendor(pk, name):
    return Vendor(pk=pk, name=name)

def create_vendor_part(part_id, vendor_id, vendor_pn, price):
    return Vendor_Part(part_id=part_id, vendor_id=vendor_id,
                       vendor_pn=vendor_pn, price=price)

def create_order(pk, vendor_id, warehouse_id):
    now = timezone.now()
    then = now + timedelta(days=30)
    return Order(pk=pk, ts=now, warehouse_id=warehouse_id,
                 expected_arrival_date=then,
                 status=Order.STATUS_OPEN, vendor_id=vendor_id)

def create_ship(pk, order_id, warehouse_id, tracking,
                cost_freight, cost_other, cost_discount):
    return Shipment(pk=pk, ts=timezone.now(), tracking=tracking,
                    ordr_id=order_id, warehouse_id=warehouse_id,
                    cost_freight=cost_freight, cost_other=cost_other,
                    cost_discount=cost_discount)

def create_inventory(pk, warehouse_id):
    return Inventory(pk=pk, ts=timezone.now(), warehouse_id=warehouse_id)

def create_assembly_item(assy_id=None, comp_id=None, qty=1):
    return Assembly_Item(assy_id=assy_id, comp_id=comp_id, qty=qty)

def create_line_item(txtn_id, part_id, qty, line_cost, index):
    return Line_Item(txtn_id=txtn_id, part_id=part_id, qty=qty,
                     line_cost=line_cost, index=index)

class Part_Method_Tests(TestCase):

    def test_is_orderable(self):
        p = Part(status=Part.STATUS_PREVIEW)
        self.assertEqual(p.is_orderable(), False)

        p = Part(status=Part.STATUS_ACTIVE)
        self.assertEqual(p.is_orderable(), True)

        p = Part(status=Part.STATUS_DEPRECATED)
        self.assertEqual(p.is_orderable(), True)

        p = Part(status=Part.STATUS_OBSOLETE)
        self.assertEqual(p.is_orderable(), False)

    def test_equivalent_parts(self):
        p1 = create_part(1)
        self.assertEqual(p1.equivalent_parts(), set([p1]))

        p2 = create_part(2)

        p1.substitute = p2
        p1.save()
        p2.save()
        self.assertEqual(p1.equivalent_parts(), set([p1, p2]))
        self.assertEqual(p2.equivalent_parts(), set([p1, p2]))

        p3 = create_part(3)
        p3.substitute = p2
        p3.save()
        self.assertEqual(p1.equivalent_parts(), set([p1, p2, p3]))
        self.assertEqual(p2.equivalent_parts(), set([p1, p2, p3]))
        self.assertEqual(p3.equivalent_parts(), set([p1, p2, p3]))

        p1.substitute = None
        p1.save()
        self.assertEqual(p1.equivalent_parts(), set([p1]))
        self.assertEqual(p2.equivalent_parts(), set([p2, p3]))
        self.assertEqual(p3.equivalent_parts(), set([p2, p3]))

        p1.delete()
        p2.delete()
        p3.delete()

class Access_Tests(TestCase):
    @classmethod
    def setup_part(cls):
        p1 = create_part(1)
        p1.save()
        ai = create_assembly_item(assy_id=1, comp_id=2, qty=3)
        ai.refdes = 'R1,R2,R3'
        ai.save()
        p = create_vendor_part(1, 2, 'ASSY_HOUSE_VENDOR_PART_#1', 325.654321)
        p.save()

    @classmethod
    def setup_vendor(cls):
        v = create_vendor(1, 'Vendor #1')
        v.save()
        p = create_vendor_part(1, 1, 'VENDOR_PART_#1', 535.123456)
        p.save()

    @classmethod
    def setup_warehouse(cls):
        w = create_warehouse(1, 'Warehouse #1')
        w.save()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        perms.add_custom_permissions(auth_models)
        u = get_user_model().objects.create_user('test', 'test@magic.com',
                                                 'magic')
        u.user_permissions.set([perms.VIEW_PERM, perms.EDIT_PERM])

        p2 = create_part(2)
        p2.save()

        p3 = create_part(3)
        p3.save()

        w = create_warehouse(2, 'Assyhouse #1')
        w.save()
        v = create_vendor(2, 'Assyhouse #1')
        v.save()

        Access_Tests.setup_part()
        Access_Tests.setup_vendor()
        Access_Tests.setup_warehouse()

    # def test_redirect_to_https(self):
    #     url = reverse('epic:epic_index')
    #     response = self.client.get(url, secure=False)
    #     self.assertEqual(response.status_code, 301)
    #     self.assertIsNotNone(re.match('https://', response.url))

    def test_unauthenticated_accesses(self):
        response = self.client.get(reverse('epic:part_info') + '?pid=1',
                                   secure=True)
        self.assertEqual(response.status_code, 302)

        for url, _ in URL_Iterator(urls.urlpatterns):
            m = re.match(r'/epic/dal/', url)
            # auto-completion URLs respond with 403 (Forbidden)
            # others return redirect to login page:
            if m:
                expected = 403
            else:
                expected = 302
            response = self.client.get(url, secure=True)
            if response.status_code != expected:
                print('Unauthenticated access to %s returned %s, expected %s' %
                      (url, response.status_code, expected))
            self.assertEqual(response.status_code, expected)

    def test_authenticated_accesses(self):
        self.client.login(username='test', password='magic')

        for url, pattern_name in URL_Iterator(urls.urlpatterns):
            if re.match('order_', pattern_name):
                # order part 1 (an assembly) from vendor 'Assyhouse #1'
                o = create_order(1, 2, 1)
                o.save()
                l = create_line_item(1, 1, qty=13, line_cost='15.95', index=1)
                l.save()
            elif re.match('ship_', pattern_name):
                # order part 1(an assembly) from vendor 'Assyhouse #1'
                o = create_order(2, 2, 1)
                o.save()
                l = create_line_item(2, 1, qty=13, line_cost='15.95', index=1)
                l.save()
                s = create_ship(1, 2, 1, '1234,5678', 1.1, 2.2, 3.3)
                s.save()
                l = create_line_item(1, 1, qty=2, line_cost='2', index=1)
                l.save()
            elif re.match('warehouse_inventory_', pattern_name):
                i = create_inventory(1, 1)
                i.save()

            expect = 200
            expect_url = None
            post_action = []
            if pattern_name == 'search_results':
                url += '?q=1'
                # with a unique result, search redirects to the detail page:
                expect = 302
                expect_url = reverse('epic:part_detail', kwargs={'pk': 1})
            elif pattern_name == 'part_delete':
                expect = 302
                expect_url = reverse('epic:part_list')
                post_action.append(Access_Tests.setup_part)
            elif pattern_name == 'ship_delete':
                expect = 302
                expect_url = reverse('epic:ship_list')
            elif pattern_name == 'vendor_delete':
                expect = 302
                expect_url = reverse('epic:vendor_list')
                post_action.append(Access_Tests.setup_vendor)
            elif pattern_name == 'warehouse_delete':
                expect = 302
                expect_url = reverse('epic:warehouse_list')
                post_action.append(Access_Tests.setup_warehouse)
            elif pattern_name == 'warehouse_inventory_delete':
                expect = 302
                expect_url = reverse('epic:warehouse_detail', kwargs={'pk':1})
            elif pattern_name == 'order_delete':
                expect = 302
                expect_url = reverse('epic:order_list')
            elif pattern_name in ['datasheet_add_part', 'datasheet_detail',
                                  'datasheet_edit', 'datasheet_delete']:
                continue

            response = self.client.get(url, secure=True)

            # we can't only have one transaction with pk=1...
            Shipment.objects.all().delete()
            Order.objects.all().delete()
            Inventory.objects.all().delete()

            if response.status_code != expect:
                print('access to %s failed' % url)
                if hasattr(response, 'url'):
                    print(' redirect URL: %s' % response.url)
            self.assertEqual(response.status_code, expect)
            if expect_url is not None:
                url = response.url
                if url[-1] == '/':
                    url = url[:-1]
                self.assertEqual(url, expect_url)

            for action in post_action:
                action()

    def test_form_upload(self):
        self.client.login(username='test', password='magic')

        # Verify that we can post item formset to an order:
        response = self.client.post('/epic/order/add',
                                    {
                                        'ts': '2018-08-06',
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'This is a note',
                                        'line_item_set-TOTAL_FORMS': 12,
                                        'line_item_set-INITIAL_FORMS': 0,
                                        'line_item_set-MIN_NUM_FORMS': 0,
                                        'line_item_set-MAX_NUM_FORMS': 1000,
                                        'line_item_set-0-qty': 8,
                                        'line_item_set-0-part': 3,
                                        'line_item_set-0-vendor_pn': 'VPN #1',
                                        'line_item_set-0-part_cost': 1,
                                        'line_item_set-0-line_cost': 120,
                                        'line_item_set-1-qty': 18,
                                        'line_item_set-1-part': 2,
                                        'line_item_set-1-vendor_pn': 'VPN #2',
                                        'line_item_set-1-part_cost': 1,
                                        'line_item_set-1-line_cost': 121
                                    })

        # response.request['wsgi.input']._FakePayload__content.seek(0)
        # print('REQUEST',
        #       response.request['wsgi.input']._FakePayload__content.read()
        #       .decode('utf-8'))
        if response.status_code != 302:
            for name in response.context.keys():
                if re.match(r'[^_]+_form$', name):
                    print('%s.errors:' % name, response.context[name].errors)

        self.assertEqual(response.status_code, 302)
        order = Order.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=order)
        self.assertEqual(line_items[0].qty, 8)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 120)
        self.assertEqual(line_items[1].qty, 18)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 121)

        # Verify that we can post an item formset to a shipment:
        response = self.client.post('/epic/order/%d/add-shipment' % order.id,
                                    {
                                        'ts': '2018-08-06',
                                        'cost_freight': 10.0,
                                        'cost_other': 3.14,
                                        'cost_discount': 2.74,
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'note 2',
                                        'line_item_set-TOTAL_FORMS': 12,
                                        'line_item_set-INITIAL_FORMS': 0,
                                        'line_item_set-MIN_NUM_FORMS': 0,
                                        'line_item_set-MAX_NUM_FORMS': 1000,
                                        'line_item_set-0-qty': 7,
                                        'line_item_set-0-part': 3,
                                        'line_item_set-0-vendor_pn': 'VPN #1',
                                        'line_item_set-0-part_cost': 1,
                                        'line_item_set-0-line_cost': 120,
                                        'line_item_set-1-qty': 17,
                                        'line_item_set-1-part': 2,
                                        'line_item_set-1-vendor_pn': 'VPN #2',
                                        'line_item_set-1-part_cost': 1,
                                        'line_item_set-1-line_cost': 121
                                    })
        self.assertEqual(response.status_code, 302)
        ship = Shipment.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=ship)
        self.assertEqual(line_items[0].qty, 7)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 120)
        self.assertEqual(line_items[1].qty, 17)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 121)

        # Verify that we can post a delta formset as an inventory:
        response = self.client.post('/epic/warehouse/1/add-inventory',
                                    {
                                        'ts': '2018-08-06 08:15 AM',
                                        'notes': 'Note #3',
                                        'delta_set-TOTAL_FORMS': 12,
                                        'delta_set-INITIAL_FORMS': 0,
                                        'delta_set-MIN_NUM_FORMS': 0,
                                        'delta_set-MAX_NUM_FORMS': 1000,
                                        'delta_set-0-adj': 6,
                                        'delta_set-0-part': 2,
                                        'delta_set-0-vendor_pn': 'VPN #2',
                                        'delta_set-0-part_cost': 1,
                                        'delta_set-0-line_cost': 120,
                                        'delta_set-1-adj': 16,
                                        'delta_set-1-part': 3,
                                        'delta_set-1-vendor_pn': 'VPN #3',
                                        'delta_set-1-part_cost': 1,
                                        'delta_set-1-line_cost': 121
                                    })
        self.assertEqual(response.status_code, 302)
        inv = Inventory.objects.all()[0]
        line_items = Delta.objects.filter(txtn=inv)
        self.assertEqual(line_items[0].part.id, 2)
        self.assertEqual(line_items[0].is_absolute, True)
        self.assertEqual(line_items[0].adj, 6)
        self.assertEqual(line_items[0].warehouse.id, 1)
        self.assertEqual(line_items[1].part.id, 3)
        self.assertEqual(line_items[1].is_absolute, True)
        self.assertEqual(line_items[1].adj, 16)
        self.assertEqual(line_items[1].warehouse.id, 1)

    def test_spreadsheet_upload(self):
        self.client.login(username='test', password='magic')
        # import pdb
        # pdb.set_trace()

        # Verify invalid column names yield the expected errors:
        response = self.client.post('/epic/order/add',
                                    {
                                        'ts': '2018-08-06',
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'This is a note',
                                        'spreadsheet': io.BytesIO(
                                            b'qty1,part1,extra,amount1\n'
                                            b'10,3,5,99\n'
                                            b'40,EP0002,,10\n'),
                                        'line_item_set-TOTAL_FORMS': 0,
                                        'line_item_set-INITIAL_FORMS': 0
                                    })
        self.assertFormError(response, 'order_form', 'spreadsheet',
                             ['Column qty missing.  It may also be called '
                              'one of: quantity, count.',
                              'Column part missing.  It may also be called '
                              'one of: pn, ep.',
                              'Column amount missing.  It may also be called '
                              'one of: cost.'])

        # Verify that we can post a CSV file to an order:
        response = self.client.post('/epic/order/add',
                                    {
                                        'ts': '2018-08-06',
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'This is a note',
                                        'spreadsheet': io.BytesIO(
                                            b'qty,part,extra,amount\n'
                                            b'10,3,5,99\n'
                                            b'40,EP0002,,10\n'),
                                        'line_item_set-TOTAL_FORMS': 0,
                                        'line_item_set-INITIAL_FORMS': 0
                                    })
        self.assertEqual(response.status_code, 302)
        order = Order.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=order)
        self.assertEqual(line_items[0].qty, 10)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 99)
        self.assertEqual(line_items[1].qty, 40)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 10)

        # Verify that we can post a CSV file to a shipment:
        response = self.client.post('/epic/order/%d/add-shipment' % order.id,
                                    {
                                        'ts': '2018-08-06',
                                        'cost_freight': 10.0,
                                        'cost_other': 3.14,
                                        'cost_discount': 2.74,
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'note 2',
                                        'spreadsheet': io.BytesIO(
                                            b'qty,part,extra,amount\n'
                                            b'5,3,5,99\n'
                                            b'10,EP0002,,10\n')
                                    })
        self.assertEqual(response.status_code, 302)
        ship = Shipment.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=ship)
        self.assertEqual(line_items[0].qty, 5)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 99)
        self.assertEqual(line_items[1].qty, 10)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 10)

        # Verify that we can post an Excel file as an inventory:
        excel_file = os.path.join(os.path.dirname(epic.__file__),
                                  'test-data', 'inventory.xlsx')
        with open(excel_file, 'rb') as f:
            response = self.client.post('/epic/warehouse/1/add-inventory',
                                        {
                                            'ts': '2018-08-06 08:15 AM',
                                            'spreadsheet': f
                                        })
        self.assertEqual(response.status_code, 302)
        inv = Inventory.objects.all()[0]
        line_items = Delta.objects.filter(txtn=inv)
        self.assertEqual(line_items[0].part.id, 2)
        self.assertEqual(line_items[0].is_absolute, True)
        self.assertEqual(line_items[0].adj, 1234)
        self.assertEqual(line_items[0].warehouse.id, 1)
        self.assertEqual(line_items[1].part.id, 3)
        self.assertEqual(line_items[1].is_absolute, True)
        self.assertEqual(line_items[1].adj, 5678)
        self.assertEqual(line_items[1].warehouse.id, 1)

    def test_order_status(self):
        '''Confirm that deleting a shipment for completed order will mark
        order as OPEN again.

        '''
        self.client.login(username='test', password='magic')
        # create order from Vendor #1:
        o = create_order(pk=4, vendor_id=1, warehouse_id=1)
        o.save()
        l = create_line_item(o.pk, 2, qty=1000, line_cost='256.35', index=1)
        l.save()

        # order should be open now
        self.assertEqual(o.status, Order.STATUS_OPEN)

        # create two shipments that should complete the order:
        s = create_ship(2, o.pk, 1, '1Z1345', 1.23, 2.34, 3.45)
        s.save()
        l = create_line_item(s.pk, 2, 900, line_cost='230.05', index=1)
        l.save()

        s = create_ship(3, o.pk, 1, '1Z1356', 1.23, 2.34, 3.45)
        s.save()
        l = create_line_item(s.pk, 2, 100, line_cost='230.05', index=1)
        l.save()

        s.finalize()

        # order should be closed now:
        o = Order.objects.get(pk=4)
        self.assertEqual(o.status, Order.STATUS_CLOSED)

        # delete the first shipment:
        url = reverse('epic:ship_delete', kwargs={'pk':2})
        response = self.client.get(url, secure=True)
        self.assertEqual(response.status_code, 302)

        # order should be open again:
        o = Order.objects.get(pk=4)
        self.assertEqual(o.status, Order.STATUS_OPEN)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-10 04:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0006_auto_20170509_0422'),
    ]

    operations = [
        migrations.AlterField(
            model_name='part',
            name='notes',
            field=models.TextField(blank=True, default='', help_text='Notes regarding this part.'),
        ),
    ]



#
#   Copyright (c) 2014-2016, 2018 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# pylint: disable=no-member
from datetime import date, datetime

from django.conf import settings as cfg
from django.contrib.auth.decorators import permission_required
from django.contrib.humanize.templatetags.humanize import intcomma
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.text import slugify

from openpyxl import Workbook
from openpyxl.styles import Font, PatternFill, fills
from openpyxl.utils import get_column_letter

from epic import perms
from epic.lib import get_stock_summary, get_inventory_summary
from epic.models import Inventory, Line_Item, Order, Part, Shipment, \
    Vendor_Part, Warehouse

FONT_NAME = 'Arial'
FONT_SIZE = 10
FMT_COUNT = '#,##0'		# integer
FMT_PRICE = '$ #,##0.000000'	# currency w/ 6 decimals
FMT_AMOUNT = '$ #,##0.00'	# currency w/ 2 decimals
# integer show negative numbers with red:
FMT_DIFF = '[Green]#,##0;[Red]-#,##0;0'

class Style():
    fields = ('font', 'fill', 'border', 'alignment', 'number_format',
              'protection')

    def __init__(self, **kwargs):
        for f in Style.fields:
            if f in kwargs:
                self.__dict__[f] = kwargs[f]

    def copy(self, **kwargs):
        ns = Style()
        for f in Style.fields:
            if f in kwargs:
                ns.__dict__[f] = kwargs[f]
            elif f in self.__dict__:
                ns.__dict__[f] = self.__dict__[f]
        return ns

    def apply(self, target):
        if 'font' in self.__dict__:
            target.font = self.font
        if 'fill' in self.__dict__:
            target.fill = self.fill
        if 'border' in self.__dict__:
            target.border = self.border
        if 'alignment' in self.__dict__:
            target.alignment = self.alignment
        if 'number_format' in self.__dict__:
            target.number_format = self.number_format
        if 'protection' in self.__dict__:
            target.protection = self.protection

class WorksheetWriter():

    def __init__(self, worksheet):
        self.worksheet = worksheet
        self.row = 1
        self.col = 0

    def set_row(self, row):
        self.row = row
        self.col = 0

    def next_row(self):
        self.set_row(self.row + 1)

    def append(self, val, style=None, col=None):
        if col is None:
            self.col += 1
            col = self.col
        else:
            self.col = col
        c = self.worksheet.cell(row=self.row, column=col)
        if val is not None:
            c.value = val
        if style is not None:
            style.apply(c)

def render_xls(filename, workbook):
    """Render the openpyxl WORKBOOK spreadsheet into an HTTP response.
    FILENAME is the name that will be given to the downloaded
    spreadsheet (sans .xlsx extension).
    """
    response = HttpResponse(content_type='application/ms-excel')
    response['Content-Disposition'] = 'attachment; filename=%s.xlsx' % filename
    workbook.save(response)
    return response

@permission_required(perms.VIEW)
def part_bom_export(request, pk):
    # pylint: disable=unused-argument
    part = get_object_or_404(Part, pk=pk)

    assy_items = part.assembly_items()
    if not assy_items:
        raise Http404	# not an assembly

    wb = Workbook()
    ws = wb.active
    # worksheet title must not be longer than 31 characters:
    ws.title = slugify(part.mfg_pn)[:31]
    ww = WorksheetWriter(ws)

    gray90 = PatternFill(fill_type=fills.FILL_SOLID, start_color='ffe5e5e5')
    gray95 = PatternFill(fill_type=fills.FILL_SOLID, start_color='fff2f2f2')

    style_normal = Style(font=Font(FONT_NAME, FONT_SIZE))
    style_comment = Style(font=Font(FONT_NAME, FONT_SIZE, italic=True))
    style_title = Style(font=Font(FONT_NAME, FONT_SIZE, bold=True))
    style_substitute = [
        Style(font=Font(FONT_NAME, FONT_SIZE), fill=gray90),
        Style(font=Font(FONT_NAME, FONT_SIZE), fill=gray95),
    ]

    cols = [('Idx', 6), ('PN', 10), ('Substitute PN', 13),
            ('Manufacturer', 24), ('Manufacturer PN', 31),
            ('Value', 8), ('Qty', 8),
            ('Price', 12), ('Refdes', 64), ('Description', 24)]
    for cspec in cols:
        ww.append(cspec[0])
        ws.column_dimensions[get_column_letter(ww.col)].width = cspec[1]
    style_title.apply(ws.row_dimensions[ww.row])
    ww.next_row()

    sub_style_idx = 0
    idx = 1
    for item in assy_items:
        best_part = item.comp.best_part()
        refdes = ', '.join(item.refdes.split(','))
        equiv_parts = best_part.equivalent_parts()

        if len(equiv_parts) > 1:
            style = style_substitute[sub_style_idx]
            sub_style_idx = (sub_style_idx + 1) % 2
        else:
            style = style_normal

        ww.append(idx)
        ww.append(str(best_part))
        ww.append(None)		# skip over Substitute PN
        ww.append(best_part.mfg)
        ww.append(best_part.mfg_pn)
        ww.append(best_part.val)
        ww.append(item.qty)
        ww.append(best_part.target_price,
                  style.copy(number_format='$ #,##0.000'))
        ww.append(refdes)
        ww.append(best_part.descr)
        style.apply(ws.row_dimensions[ww.row])
        ww.next_row()

        # add substitute rows, if any:
        for sub in best_part.equivalent_parts():
            if sub == best_part or sub.status == Part.STATUS_PREVIEW:
                continue
            ww.append(str(sub), style, col=3)
            ww.append(sub.mfg, style)
            ww.append(sub.mfg_pn, style)
            ww.append(sub.val, style)
            ww.append(None, style)	# skip over Qty
            ww.append(None, style)	# skip over Price
            ww.append(None, style)	# skip over Refdes
            ww.append(sub.descr, style)
            ww.next_row()
        idx += 1
    ww.next_row()
    ww.append('End of BOM', style_comment)
    return render_xls(slugify(ws.title), wb)

@permission_required(perms.VIEW)
def order_export(request, pk):
    # pylint: disable=unused-argument
    order = get_object_or_404(Order, pk=pk)
    line_items = Line_Item.objects.filter(txtn_id=pk)

    wb = Workbook()
    ws = wb.active
    ws.title = 'PO# %d' % order.id
    ww = WorksheetWriter(ws)

    style_normal = Style(font=Font(FONT_NAME, FONT_SIZE))
    style_comment = Style(font=Font(FONT_NAME, FONT_SIZE, italic=True))
    style_title = Style(font=Font(FONT_NAME, FONT_SIZE, bold=True))
    style_qty = style_normal.copy(number_format=FMT_COUNT)
    style_part_cost = style_normal.copy(number_format=FMT_PRICE)
    style_line_cost = style_normal.copy(number_format=FMT_AMOUNT)

    ww.append('Purchase Order #', style_title)
    ww.append('%d' % order.id, style_normal)
    ww.next_row()
    ww.append('Vendor', style_title)
    ww.append(order.vendor.name, style_normal)
    ww.next_row()
    ww.append('Order Date', style_title)
    ww.append(order.ts.strftime('%b %d, %Y'), style_normal)
    ww.next_row()
    ww.append('Expected Arrival Date', style_title)
    ww.append(order.expected_arrival_date.strftime('%b %d, %Y'), style_normal)
    ww.next_row()
    ww.next_row()

    ww.append('Bill To', style_title)
    for l in cfg.EPIC_BILL_TO_ADDRESS.split('\n'):
        ww.append(l, style_normal, col=2)
        ww.next_row()
    ww.next_row()

    ww.append('Ship To', style_title)
    for l in order.warehouse.address.split('\n'):
        ww.append(l, style_normal, col=2)
        ww.next_row()
    ww.next_row()

    ww.append('Ship By', style_title)
    ww.append(cfg.EPIC_SHIPPING_TYPE, style_normal)
    ww.next_row()

    ww.append('Shipping Account', style_title)
    ww.append(cfg.EPIC_SHIPPING_ACCOUNT, style_normal)
    ww.next_row()
    ww.next_row()

    cols = [('Idx', 20), ('Qty', 20), ('PN', 10), ('Vendor PN', 31),
            ('Manufacturer', 24), ('Manufacturer PN', 31),
            ('Price', 16), ('Amount', 16)]
    for cspec in cols:
        ww.append(cspec[0])
        ws.column_dimensions[get_column_letter(ww.col)].width = cspec[1]
    style_title.apply(ws.row_dimensions[ww.row])
    ww.next_row()

    total = 0
    for item in line_items:
        vendor_part = Vendor_Part.get(item.part.id, order.vendor.id)
        total += item.line_cost
        ww.append(item.index)
        ww.append(item.qty, style_qty)
        ww.append(str(item.part))
        ww.append(vendor_part.vendor_pn)
        ww.append(item.part.mfg)
        ww.append(item.part.mfg_pn)
        ww.append(item.part_cost(), style_part_cost)
        ww.append(item.line_cost, style_line_cost)
        style_normal.apply(ws.row_dimensions[ww.row])
        ww.next_row()
    ww.append('Order Total', style_title, col=7)
    ww.append(total, style_title.copy(number_format=FMT_AMOUNT))
    ww.next_row()
    ww.append('End of Order', style_comment)
    return render_xls('order-%d-%s' % (order.id, slugify(order.vendor.name)),
                      wb)

@permission_required(perms.VIEW)
def ship_export(request, pk):
    # pylint: disable=unused-argument
    ship = get_object_or_404(Shipment, pk=pk)
    line_items = Line_Item.objects.filter(txtn_id=pk)

    wb = Workbook()
    ws = wb.active
    ws.title = 'SO# %d' % ship.id
    ww = WorksheetWriter(ws)

    style_normal = Style(font=Font(FONT_NAME, FONT_SIZE))
    style_comment = Style(font=Font(FONT_NAME, FONT_SIZE, italic=True))
    style_title = Style(font=Font(FONT_NAME, FONT_SIZE, bold=True))
    style_qty = style_normal.copy(number_format=FMT_COUNT)

    ww.append('Ship Order #', style_title)
    ww.append('%d' % ship.id, style_normal)
    ww.next_row()

    ww.append('Tracking #', style_title)
    ww.append(ship.tracking, style_normal)
    ww.next_row()

    ww.append('Ship Date', style_title)
    ww.append(ship.ts.strftime('%b %d, %Y'), style_normal)
    ww.next_row()
    ww.append('Ship From', style_title)
    ww.append(ship.ordr.vendor.name if ship.ordr
              else ship.from_warehouse.name, style_normal)
    ww.next_row()
    ww.next_row()

    ww.append('Ship To', style_title)
    for l in ship.warehouse.address.split('\n'):
        ww.append(l, style_normal, col=2)
        ww.next_row()
    ww.next_row()

    # Don't include any price information in the ship order.  That info
    # is somewhat sensitive and kept best between the vendor and ourselves.
    # We don't want to share that with the assembly houses.
    cols = [('Idx', 20), ('Qty', 20), ('PN', 10), ('Vendor PN', 31),
            ('Manufacturer', 24), ('Manufacturer PN', 31)]
    for cspec in cols:
        if ship.ordr is None and cspec[0] == 'Vendor PN':
            continue	# no vendor for inter-warehouse shipments
        ww.append(cspec[0])
        ws.column_dimensions[get_column_letter(ww.col)].width = cspec[1]
    style_title.apply(ws.row_dimensions[ww.row])
    ww.next_row()

    for item in line_items:
        ww.append(item.index)
        ww.append(item.qty, style_qty)
        ww.append(str(item.part))
        if ship.ordr:
            vendor_part = Vendor_Part.get(item.part.id, ship.ordr.vendor.id)
            ww.append(vendor_part.vendor_pn)
        ww.append(item.part.mfg)
        ww.append(item.part.mfg_pn)
        style_normal.apply(ws.row_dimensions[ww.row])
        ww.next_row()
    ww.next_row()
    ww.append('End of Shipment', style_comment)
    return render_xls('ship-%d-%s' % (ship.id, slugify(ship.warehouse.name)),
                      wb)

@permission_required(perms.VIEW)
def warehouse_stock_export(request, pk):

    def part_ref(part_list, index):
        result = str(part_list[index][0])
        if len(part_list) > 1:
            result += ' * %s' % intcomma(part_list[index][1])
        return result

    def row_style(style, index):
        if index % 2 == 0:
            return style.copy(fill=gray95)
        return style.copy(fill=gray90)

    if 'as_of_date' in request.GET:
        as_of_date = datetime.strptime(request.GET['as_of_date'], '%Y-%m-%d') \
                             .date()
    else:
        as_of_date = date.today()

    if pk is None:
        warehouse = None
    else:
        warehouse = get_object_or_404(Warehouse, pk=pk)
    stock, stock_total = get_stock_summary(warehouse, as_of_date=as_of_date)

    wb = Workbook()
    ws = wb.active
    ws.title = 'Stock Summary '
    ww = WorksheetWriter(ws)

    gray90 = PatternFill(fill_type=fills.FILL_SOLID, start_color='ffe5e5e5')
    gray95 = PatternFill(fill_type=fills.FILL_SOLID, start_color='fff2f2f2')
    style_normal = Style(font=Font(FONT_NAME, FONT_SIZE))
    style_comment = Style(font=Font(FONT_NAME, FONT_SIZE, italic=True))
    style_title = Style(font=Font(FONT_NAME, FONT_SIZE, bold=True))
    style_qty = style_normal.copy(number_format=FMT_COUNT)

    title = 'Stock as of %s ' % as_of_date.strftime('%b %d, %Y')
    if pk is None:
        title += 'for all warehouses together'
    else:
        title += 'at warehouse %s' % warehouse.name

    ww.append(title, style_title)
    ww.next_row()
    ww.next_row()

    cols = [('Idx', 6), ('Qty', 10), ('PN', 15),
            ('Manufacturer', 24), ('Manufacturer PN', 31),
            ('Price', 16), ('Amount', 16)]
    for cspec in cols:
        ww.append(cspec[0])
        ws.column_dimensions[get_column_letter(ww.col)].width = cspec[1]
    style_title.apply(ws.row_dimensions[ww.row])
    ww.next_row()

    index = 1
    for row in stock:
        style = row_style(style_normal, index)
        ww.append(index, style)
        ww.append(row['total_qty'], row_style(style_qty, index))
        part_list = row['part_list']
        part = part_list[0][0]
        ww.append(part_ref(part_list, 0), style)
        ww.append(part.mfg, style)
        ww.append(part.mfg_pn, style)
        ww.append(row['price'], style.copy(number_format=FMT_PRICE))
        ww.append(row['amount'], style.copy(number_format=FMT_AMOUNT))
        for i in range(1, len(part_list)):
            ww.next_row()
            part = part_list[i][0]
            ww.append(None, style)
            ww.append(None, style)
            ww.append(part_ref(part_list, i), style)
            ww.append(part.mfg, style)
            ww.append(part.mfg_pn, style)
            ww.append(None, style)
            ww.append(None, style)
        ww.next_row()
        index += 1
    ww.append('Parts Total', style_title, col=6)
    ww.append(stock_total, style_title.copy(number_format=FMT_AMOUNT))
    ww.next_row()
    ww.append('End of Report', style_comment)

    return render_xls('stock-%s' % (slugify(warehouse.name) if warehouse
                                    else 'all'),
                      wb)

@permission_required(perms.VIEW)
def warehouse_stock_all_export(request):
    return warehouse_stock_export(request, None)

@permission_required(perms.VIEW)
def warehouse_inv_export(request, warehouse, pk):
    # pylint: disable=unused-argument

    def row_style(style, index):
        if index % 2 == 0:
            return style.copy(fill=gray95)
        return style.copy(fill=gray90)

    inventory = get_object_or_404(Inventory, pk=pk)
    warehouse = get_object_or_404(Warehouse, pk=warehouse)
    if inventory.warehouse_id != warehouse.id:
        raise Http404

    ret = get_inventory_summary(warehouse.id, inventory)
    inv_items = ret[0]

    wb = Workbook()
    ws = wb.active
    ws.title = 'Inventory # %d ' % inventory.id
    ww = WorksheetWriter(ws)

    gray90 = PatternFill(fill_type=fills.FILL_SOLID, start_color='ffe5e5e5')
    gray95 = PatternFill(fill_type=fills.FILL_SOLID, start_color='fff2f2f2')
    style_normal = Style(font=Font(FONT_NAME, FONT_SIZE))
    style_comment = Style(font=Font(FONT_NAME, FONT_SIZE, italic=True))
    style_title = Style(font=Font(FONT_NAME, FONT_SIZE, bold=True))
    style_qty = style_normal.copy(number_format=FMT_COUNT)

    title = 'Inventory as of %s for warehouse %s' \
            % (inventory.ts.strftime('%b %d, %Y'), warehouse.name)

    ww.append(title, style_title)
    ww.next_row()
    ww.next_row()

    # Don't include value change so this spreadsheet can be shared with
    # assembly-houses:
    cols = [('Idx', 6), ('PN', 15),
            ('Manufacturer', 24), ('Manufacturer PN', 31),
            ('New Qty', 10), ('Old Qty', 10), ('Qty Change', 10),
            ('Net Change', 10)]
    for cspec in cols:
        ww.append(cspec[0])
        ws.column_dimensions[get_column_letter(ww.col)].width = cspec[1]
    style_title.apply(ws.row_dimensions[ww.row])
    ww.next_row()

    index = 1
    for row in inv_items:
        style = row_style(style_normal, index)
        ww.append(index, style)
        qstyle = row_style(style_qty, index)
        net_dif = 0
        for i in range(len(row['qty_dif_list'])):
            net_dif += row['qty_dif_list'][i][0]
        for i in range(len(row['qty_new_list'])):
            if i > 0:
                ww.next_row()
                ww.append('', style)	# write empty index column
            part = row['qty_new_list'][i][1]
            ww.append(str(part), style)
            ww.append(part.mfg)
            ww.append(part.mfg_pn)
            ww.append(row['qty_new_list'][i][0], qstyle)
            ww.append(row['qty_old_list'][i][0], qstyle)
            ww.append(row['qty_dif_list'][i][0], qstyle)
            if i == 0:
                ww.append(net_dif, qstyle)
            else:
                ww.append('', qstyle)
        ww.next_row()
        index += 1
    ww.next_row()
    ww.append('End of Inventory', style_comment)

    return render_xls('inventory-%d-%s' % (inventory.id,
                                           slugify(warehouse.name)),
                      wb)
